/* vi: set sw=4 ts=4: */
/*  nc: mini-netcat - built from the ground up for LRP
    Copyright (C) 1998  Charles P. Wright

    0.0.1   6K      It works.
    0.0.2   5K      Smaller and you can also check the exit condition if you wish.
    0.0.3           Uses select()

    19980918 Busy Boxed! Dave Cinege
    19990512 Uses Select. Charles P. Wright
    19990513 Fixes stdin stupidity and uses buffers.  Charles P. Wright

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>
#include <signal.h>

#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>
#include <arpa/inet.h>
#include <netdb.h>
#include <sys/time.h>
#include <sys/ioctl.h>
#include "busybox.h"

static void timeout(int signum)
{
	bb_error_msg_and_die("Timed out");
}

static int set_socket_options(int sfd, int sock_af
#ifdef CONFIG_NC_QOS
			      , int qos_tos
#endif
			      )
{
	int on = 1;

	if (setsockopt(sfd, SOL_SOCKET, SO_REUSEADDR, &on, sizeof(on)) == -1)
		return -1;
#ifdef CONFIG_NC_QOS
	/* qos */
	switch (sock_af) {
		case AF_INET:
		{
			if (setsockopt(sfd, SOL_IP, IP_TOS,
				       &qos_tos, sizeof(qos_tos)) < 0)
				return -1;
		}
		break;
#ifdef IPV6_TCLASS
		case AF_INET6:
		{
			if (setsockopt(sfd, SOL_IPV6, IPV6_TCLASS,
				       &qos_tos, sizeof(qos_tos)) < 0)
				return -1;
		}
		break;
#endif /* IPV6_TCLASS */
		default:
		break;
	}
#endif /* CONFIG_NC_QOS */
	return 0;
}

static int create_socket(int af, int type, int flags, char *host,
			 char *dport, char *sport
#ifdef CONFIG_NC_QOS
			 , int qos_tos
#endif
			 )
{
	struct addrinfo hints, *res, *aip;
	int sockfd = -1;
	int r;

	memset(&hints, 0, sizeof(hints));
	hints.ai_flags    = flags;
	hints.ai_family   = af;
	hints.ai_socktype = type;

	if ((r = getaddrinfo(host, dport, &hints, &res)) != 0)
		bb_error_msg_and_die(gai_strerror(r));

	for (aip = res; aip; aip = aip->ai_next) {
#ifdef CONFIG_FEATURE_IPV6
		int on = 1;
#endif
		sockfd = socket(aip->ai_family,
				aip->ai_socktype,
				aip->ai_protocol);

		if (sockfd < 0)
			continue;
#ifdef CONFIG_NC_QOS
		if (set_socket_options(sockfd, aip->ai_family, qos_tos) < 0)
			goto err;
#endif
		if (flags & AI_PASSIVE) {
			int tmpfd;
#ifdef CONFIG_FEATURE_IPV6
			if (aip->ai_family == PF_INET6)
				setsockopt(sockfd, IPPROTO_IPV6, IPV6_V6ONLY,
					   (char *)&on, sizeof(on));
#endif
			if (bind(sockfd, aip->ai_addr, aip->ai_addrlen) != 0)
				goto err;

#ifdef CONFIG_NC_UDP
			if (aip->ai_socktype == SOCK_DGRAM)
				break;
#endif

			if (listen(sockfd, SOMAXCONN) != 0)
				goto err;

			if ((tmpfd = accept(sockfd, NULL, NULL)) < 0)
				goto err;

			close(sockfd);
			sockfd = tmpfd;
			break;
		} else {
			if (sport) {
				int lport;
				struct sockaddr_storage ss;

				memset(&ss, 0, sizeof ss);

				lport = bb_lookup_port(sport, "tcp", 0);
				ss.ss_family = aip->ai_family;
#ifdef CONFIG_FEATURE_IPV6
				if (aip->ai_family == AF_INET6)
					((struct sockaddr_in6 *)&ss)->sin6_port = lport;
				else
#endif
					((struct sockaddr_in *)&ss)->sin_port = lport;

				if (bind(sockfd, (struct sockaddr *)&ss, sizeof ss) != 0)
					goto err;
			}

			if (connect(sockfd, aip->ai_addr, aip->ai_addrlen) != 0)
				goto err;
			break;
		}
		continue;
	  err:
		close(sockfd);
		sockfd = -1;
	}

	if (sockfd == -1)
		bb_perror_msg_and_die("socket");

	freeaddrinfo(res);
	return sockfd;
}

int nc_main(int argc, char **argv)
{
	int do_listen = 0, delay = 0, wsecs = 0, opt, sfd;
#ifdef CONFIG_NC_QOS
	int qos_tos = 0, qos_prio = 0;
#endif

#define buf bb_common_bufsiz1

#ifdef CONFIG_NC_GAPING_SECURITY_HOLE
	char *pr00gie = NULL;
#endif

	char *sport = NULL;
	int sock_af = PF_UNSPEC;
	int sock_type = SOCK_STREAM;
	fd_set readfds, testfds;

	while ((opt = getopt(argc, argv,
			     "lp:i:w:"
#ifdef CONFIG_NC_GAPING_SECURITY_HOLE
			     "e:"
#endif
#if CONFIG_FEATURE_IPV6
			     "46"
#endif
#ifdef CONFIG_NC_UDP
			     "u"
#endif
#ifdef CONFIG_NC_QOS
			     "D:P:"
#endif
			     )) > 0) {
		switch (opt) {
			case 'l':
				do_listen++;
				break;
			case 'p':
				sport = optarg;
				break;
			case 'i':
				delay = atoi(optarg);
				break;
			case 'w':
				wsecs = atoi(optarg);
				break;
#ifdef CONFIG_NC_GAPING_SECURITY_HOLE
			case 'e':
				pr00gie = optarg;
				break;
#endif
#if CONFIG_FEATURE_IPV6
			case '4':
				sock_af = PF_INET;
				break;
			case '6':
				sock_af = PF_INET6;
				break;
#endif
#ifdef CONFIG_NC_UDP
			case 'u':
				sock_type = SOCK_DGRAM;
				break;
#endif
#ifdef CONFIG_NC_QOS
			case 'D':
				qos_tos = atoi(optarg);
				qos_tos = (qos_tos & 0x3f) << 2;
				break;
			case 'P':
				qos_prio = atoi(optarg);
				break;
#endif
			default:
				bb_show_usage();
		}
	}

	if ((do_listen && optind != argc) ||
	    (!do_listen && optind + 2 != argc))
		bb_show_usage();

	if (wsecs) {
		signal(SIGALRM, timeout);
		alarm(wsecs);
	}

	if (do_listen)
		sfd = create_socket(sock_af == PF_UNSPEC ? PF_INET : sock_af,
				    sock_type, AI_PASSIVE,
				    NULL, sport, NULL
#ifdef CONFIG_NC_QOS
				    ,qos_tos
#endif
				    );
	else
		sfd = create_socket(sock_af,
				    sock_type, 0,
				    argv[optind], argv[optind + 1], sport
#ifdef CONFIG_NC_QOS
				    , qos_tos
#endif
				    );
#ifdef CONFIG_NC_QOS
	if (setsockopt(sfd, SOL_SOCKET, SO_PRIORITY,
		       &qos_prio, sizeof(qos_prio)) < 0)
		bb_perror_msg_and_die("set user-priority field failed");
#endif
	if (wsecs) {
		alarm(0);
		signal(SIGALRM, SIG_DFL);
	}

#ifdef CONFIG_NC_GAPING_SECURITY_HOLE
	/* -e given? */
	if (pr00gie) {
		dup2(sfd, 0);
		close(sfd);
		dup2(0, 1);
		dup2(0, 2);
		execl(pr00gie, pr00gie, NULL);
		/* Don't print stuff or it will go over the wire.... */
		_exit(-1);
	}
#endif /* CONFIG_NC_GAPING_SECURITY_HOLE */

	FD_ZERO(&readfds);
	FD_SET(sfd, &readfds);
	FD_SET(STDIN_FILENO, &readfds);

	while (1) {
#ifdef CONFIG_NC_UDP
		struct sockaddr_storage ss;
		socklen_t slen = sizeof ss;
		int connected = 0;
#endif
		int fd;
		int ofd;
		int nread;

		testfds = readfds;

		if (select(FD_SETSIZE, &testfds, NULL, NULL, NULL) < 0)
			bb_perror_msg_and_die("select");

		for (fd = 0; fd < FD_SETSIZE; fd++) {
			if (FD_ISSET(fd, &testfds)) {
#ifdef CONFIG_NC_UDP
				if (fd == sfd) {
					if ((nread = recvfrom(fd,
							      buf,
							      sizeof(buf),
							      0,
							      (struct sockaddr *)&ss,
							      &slen)) < 0)
						bb_perror_msg_and_die("recvfrom");

					if ((sock_type == SOCK_DGRAM) &&
					    !connected) {
						if (connect(sfd,
							    (struct sockaddr *)&ss,
							    sizeof ss))
							bb_perror_msg_and_die("connect");
						connected++;
					}
				}
				else
#endif
					if ((nread = safe_read(fd,
							       buf,
							       sizeof(buf))) < 0)
						bb_perror_msg_and_die(bb_msg_read_error);

				if (fd == sfd) {
					if (nread == 0)
						exit(0);
					ofd = STDOUT_FILENO;
				} else {
					if (nread <= 0) {
						shutdown(sfd, 1 /* send */ );
						close(STDIN_FILENO);
						FD_CLR(STDIN_FILENO, &readfds);
					}
					ofd = sfd;
				}

				if (bb_full_write(ofd, buf, nread) < 0)
					bb_perror_msg_and_die(bb_msg_write_error);

				if (delay > 0)
					sleep(delay);
			}
		}
	}
}
